WHENEVER SQLERROR EXIT SQL.SQLCODE

SET DEFINE ON

ALTER SESSION SET CURRENT_SCHEMA=&_vUsername
/

--- CLEAN

SET TERMOUT OFF
WHENEVER SQLERROR CONTINUE
DROP TABLE FACKEYS PURGE
/
DROP TABLE BFORCEPROT
/
DROP TABLE TRUSTEDLOCS
/
DROP CONTEXT TWOFACTOR_CTX
/
WHENEVER SQLERROR EXIT SQL.SQLCODE
SET TERMOUT ON

--- TABLE FACKEYS

CREATE TABLE FACKEYS
(
    USERNAME            VARCHAR2(30 CHAR)                         NOT NULL,
    KEY                 RAW(24)                                   NOT NULL,
    SECURE_KEY          VARCHAR2(32 CHAR) DEFAULT 'jIqIcHbVqdpj7b4PzBpM2DiIeBvmHFV/' NOT NULL,
    CYCLE_TIME          VARCHAR2(03 CHAR) DEFAULT '30'            NOT NULL,
    CORR_TIME           VARCHAR2(05 CHAR) DEFAULT '0'             NOT NULL,
    KEY_METHOD          VARCHAR2(10 CHAR) DEFAULT 'app512'        NOT NULL,
    STATUS              VARCHAR2(08 CHAR) DEFAULT 'DISABLED'      NOT NULL,
    VALIDATED           VARCHAR2(13 CHAR) DEFAULT 'NOT VALIDATED' NOT NULL,
    TIME_GAP            NUMBER(4,0)       DEFAULT 240             NOT NULL,
    MAX_TRUSTED_SOURCES NUMBER(2,0)       DEFAULT 5               NOT NULL
) TABLESPACE SYSAUX
/


BEGIN
    IF  &_O_RELEASE LIKE '12%'
    THEN
        EXECUTE IMMEDIATE 'ALTER TABLE FACKEYS MODIFY KEY INVISIBLE';
    END IF;
END;
/

COMMENT ON TABLE  FACKEYS                     IS '2-Factor Auth control table.'
/
COMMENT ON COLUMN FACKEYS.USERNAME            IS 'User schema name.'
/
COMMENT ON COLUMN FACKEYS.KEY                 IS 'Encrypted user secret.'
/
COMMENT ON COLUMN FACKEYS.SECURE_KEY          IS 'Encrypted user secret.'
/
COMMENT ON COLUMN FACKEYS.STATUS              IS 'Current user status: ENABLED/DISABLED.'
/
COMMENT ON COLUMN FACKEYS.VALIDATED           IS 'Has user ever validated his code?'
/
COMMENT ON COLUMN FACKEYS.TIME_GAP            IS 'Time difference in secs to avoid clock errors.'
/
COMMENT ON COLUMN FACKEYS.MAX_TRUSTED_SOURCES IS 'Max number of trusted sources.';

ALTER TABLE FACKEYS ADD CONSTRAINT CK_STATUS CHECK (STATUS IN ('DISABLED','ENABLED'))
/
ALTER TABLE FACKEYS ADD CONSTRAINT CK_VALIDATED CHECK (VALIDATED IN ('NOT VALIDATED','VALIDATED'))
/
ALTER TABLE FACKEYS ADD CONSTRAINT CK_STAT_VALID CHECK (NOT(VALIDATED = 'NOT VALIDATED' AND STATUS = 'ENABLED'))
/
ALTER TABLE FACKEYS ADD CONSTRAINT CK_TIME_GAP CHECK (TIME_GAP <= 600) -- Clock error. Max is ±10min = 20 min
/
ALTER TABLE FACKEYS ADD CONSTRAINT PK_FACKEYS PRIMARY KEY (USERNAME) USING INDEX TABLESPACE SYSAUX
/

--- TABLE BFORCEPROT

CREATE TABLE BFORCEPROT
(
    USERNAME      VARCHAR2(30 CHAR)            NOT NULL,
    MAX_TRIES     NUMBER(2,0)       DEFAULT 3  NOT NULL,
    WAIT_DURATION NUMBER(3,0)       DEFAULT 30 NOT NULL,
    FIRST_FAILED  TIMESTAMP(6) WITH TIME ZONE      NULL,
    TRIES         NUMBER(2,0)                      NULL
) TABLESPACE SYSAUX
/

COMMENT ON TABLE  BFORCEPROT               IS 'Control user tries to avoid B. Force attacks.'
/
COMMENT ON COLUMN BFORCEPROT.USERNAME      IS 'User schema name.'
/
COMMENT ON COLUMN BFORCEPROT.MAX_TRIES     IS 'Max tries attempts allowed on "WAIT_DURATION" window.'
/
COMMENT ON COLUMN BFORCEPROT.WAIT_DURATION IS 'Number of seconds that "MAX_TRIES" are valid.'
/
COMMENT ON COLUMN BFORCEPROT.FIRST_FAILED  IS 'First wrong attempt on this "WAIT_DURATION" window.'
/
COMMENT ON COLUMN BFORCEPROT.TRIES         IS 'Number of guesses already tried in this "WAIT_DURATION" window.'
/

ALTER TABLE BFORCEPROT ADD CONSTRAINT CK_FAILED_TRIES CHECK ((FIRST_FAILED IS NULL AND TRIES IS NULL) OR (FIRST_FAILED IS NOT NULL AND TRIES IS NOT NULL))
/
ALTER TABLE BFORCEPROT ADD CONSTRAINT PK_BFORCEPROT PRIMARY KEY (USERNAME) USING INDEX TABLESPACE SYSAUX
/
ALTER TABLE BFORCEPROT ADD CONSTRAINT FK_BFORCEPROT FOREIGN KEY (USERNAME) REFERENCES FACKEYS (USERNAME) ON DELETE CASCADE
/

--- TABLE TRUSTEDLOCS

CREATE TABLE TRUSTEDLOCS
(
    USERNAME    VARCHAR2(30  CHAR)          NOT NULL,
    OSUSER      VARCHAR2(30  CHAR)          NOT NULL,
    MACHINE     VARCHAR2(64  CHAR)          NOT NULL,
    TERMINAL    VARCHAR2(30  CHAR)              NULL,
    PROGRAM     VARCHAR2(48  CHAR)          NOT NULL,
    IP_ADDRESS  VARCHAR2(256 CHAR)              NULL,
    START_DATE  TIMESTAMP(6) WITH TIME ZONE NOT NULL,
    END_DATE    TIMESTAMP(6) WITH TIME ZONE NOT NULL
) TABLESPACE SYSAUX
/

COMMENT ON TABLE  TRUSTEDLOCS            IS 'Control user trusted locations to automatically authenticate.'
/
COMMENT ON COLUMN TRUSTEDLOCS.USERNAME   IS 'User schema name.'
/
COMMENT ON COLUMN TRUSTEDLOCS.OSUSER     IS 'OS User from trusted location.'
/
COMMENT ON COLUMN TRUSTEDLOCS.MACHINE    IS 'Machine from trusted location.'
/
COMMENT ON COLUMN TRUSTEDLOCS.TERMINAL   IS 'Terminal from trusted location.'
/
COMMENT ON COLUMN TRUSTEDLOCS.PROGRAM    IS 'Program from trusted location.'
/
COMMENT ON COLUMN TRUSTEDLOCS.IP_ADDRESS IS 'IP Address from trusted location.'
/
COMMENT ON COLUMN TRUSTEDLOCS.START_DATE IS 'Since when this location is valid.'
/
COMMENT ON COLUMN TRUSTEDLOCS.END_DATE   IS 'Until when this location is valid.'
/


ALTER TABLE TRUSTEDLOCS ADD CONSTRAINT UK_TRUSTEDLOCS UNIQUE (USERNAME, OSUSER, MACHINE, TERMINAL, PROGRAM, IP_ADDRESS) USING INDEX TABLESPACE SYSAUX
/
ALTER TABLE TRUSTEDLOCS ADD CONSTRAINT FK_TRUSTEDLOCS FOREIGN KEY (USERNAME) REFERENCES FACKEYS (USERNAME) ON DELETE CASCADE
/

--- CONTEXT TWOFACTOR_CTX

CREATE CONTEXT TWOFACTOR_CTX USING TWOFACTOR_INTERNAL
/
